/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package lb

import (
	"context"
	"strconv"
	"strings"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/validation"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudapi/lb"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func resourceLBBackendCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceLBBackendCreate")

	haveLBID, err := existLBID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	if !haveLBID {
		return diag.Errorf("resourceLBBackendCreate: can't create LB backend because LBID %d is not allowed or does not exist", d.Get("lb_id").(int))
	}

	c := m.(*controller.ControllerCfg)
	req := lb.BackendCreateRequest{}

	req.BackendName = d.Get("name").(string)
	req.LBID = uint64(d.Get("lb_id").(int))

	if algorithm, ok := d.GetOk("algorithm"); ok {
		req.Algorithm = algorithm.(string)
	}
	if inter, ok := d.GetOk("inter"); ok {
		req.Inter = uint64(inter.(int))
	}
	if downinter, ok := d.GetOk("downinter"); ok {
		req.DownInter = uint64(downinter.(int))
	}
	if rise, ok := d.GetOk("rise"); ok {
		req.Rise = uint64(rise.(int))
	}
	if fall, ok := d.GetOk("fall"); ok {
		req.Fall = uint64(fall.(int))
	}
	if slowstart, ok := d.GetOk("slowstart"); ok {
		req.SlowStart = uint64(slowstart.(int))
	}
	if maxconn, ok := d.GetOk("maxconn"); ok {
		req.MaxConn = uint64(maxconn.(int))
	}
	if maxqueue, ok := d.GetOk("maxqueue"); ok {
		req.MaxQueue = uint64(maxqueue.(int))
	}
	if weight, ok := d.GetOk("weight"); ok {
		req.Weight = uint64(weight.(int))
	}

	_, err = c.CloudAPI().LB().BackendCreate(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.Itoa(d.Get("lb_id").(int)) + "#" + d.Get("name").(string))

	_, err = utilityLBBackendCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	return resourceLBBackendRead(ctx, d, m)
}

func resourceLBBackendRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceLBBackendRead")

	b, err := utilityLBBackendCheckPresence(ctx, d, m)
	if b == nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	lbId, _ := strconv.ParseInt(strings.Split(d.Id(), "#")[0], 10, 32)

	flattenResourceLBBackend(d, b, lbId)

	return nil
}

func resourceLBBackendDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceLBBackendDelete")

	_, err := utilityLBBackendCheckPresence(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	c := m.(*controller.ControllerCfg)
	req := lb.BackendDeleteRequest{
		LBID:        uint64(d.Get("lb_id").(int)),
		BackendName: d.Get("name").(string),
	}

	_, err = c.CloudAPI().LB().BackendDelete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceLBBackendUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceLBBackendEdit")
	c := m.(*controller.ControllerCfg)

	haveLBID, err := existLBID(ctx, d, m)
	if err != nil {
		return diag.FromErr(err)
	}

	if !haveLBID {
		return diag.Errorf("resourceLBBackendUpdate: can't update LB backend because LBID %d is not allowed or does not exist", d.Get("lb_id").(int))
	}

	req := lb.BackendUpdateRequest{
		LBID:        uint64(d.Get("lb_id").(int)),
		BackendName: d.Get("name").(string),
	}

	if d.HasChange("algorithm") {
		req.Algorithm = d.Get("algorithm").(string)
	}
	if d.HasChange("inter") {
		req.Inter = uint64(d.Get("inter").(int))
	}
	if d.HasChange("downinter") {
		req.DownInter = uint64(d.Get("downinter").(int))
	}
	if d.HasChange("rise") {
		req.Rise = uint64(d.Get("rise").(int))
	}
	if d.HasChange("fall") {
		req.Fall = uint64(d.Get("fall").(int))
	}
	if d.HasChange("slowstart") {
		req.SlowStart = uint64(d.Get("slowstart").(int))
	}
	if d.HasChange("maxconn") {
		req.MaxConn = uint64(d.Get("maxconn").(int))
	}
	if d.HasChange("maxqueue") {
		req.MaxQueue = uint64(d.Get("maxqueue").(int))
	}
	if d.HasChange("weight") {
		req.Weight = uint64(d.Get("weight").(int))
	}

	_, err = c.CloudAPI().LB().BackendUpdate(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	//TODO: перенести servers сюда

	return resourceLBBackendRead(ctx, d, m)
}

func ResourceLBBackend() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceLBBackendCreate,
		ReadContext:   resourceLBBackendRead,
		UpdateContext: resourceLBBackendUpdate,
		DeleteContext: resourceLBBackendDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout600s,
			Read:    &constants.Timeout300s,
			Update:  &constants.Timeout300s,
			Delete:  &constants.Timeout300s,
			Default: &constants.Timeout300s,
		},

		Schema: map[string]*schema.Schema{
			"lb_id": {
				Type:        schema.TypeInt,
				Required:    true,
				Description: "ID of the LB instance to backendCreate",
			},
			"name": {
				Type:        schema.TypeString,
				Required:    true,
				Description: "Must be unique among all backends of this LB - name of the new backend to create",
			},
			"algorithm": {
				Type:         schema.TypeString,
				Optional:     true,
				Computed:     true,
				ValidateFunc: validation.StringInSlice([]string{"roundrobin", "static-rr", "leastconn"}, false),
			},
			"guid": {
				Type:     schema.TypeString,
				Computed: true,
			},
			"downinter": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"fall": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"inter": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"maxconn": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"maxqueue": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"rise": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"slowstart": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"weight": {
				Type:     schema.TypeInt,
				Optional: true,
				Computed: true,
			},
			"servers": {
				Type:     schema.TypeList,
				Optional: true,
				Computed: true,
				Elem: &schema.Resource{
					Schema: map[string]*schema.Schema{
						"address": {
							Type:     schema.TypeString,
							Optional: true,
							Computed: true,
						},
						"check": {
							Type:     schema.TypeString,
							Optional: true,
							Computed: true,
						},
						"guid": {
							Type:     schema.TypeString,
							Computed: true,
						},
						"name": {
							Type:     schema.TypeString,
							Optional: true,
							Computed: true,
						},
						"port": {
							Type:     schema.TypeInt,
							Optional: true,
							Computed: true,
						},
						"server_settings": {
							Type:     schema.TypeList,
							Optional: true,
							Computed: true,
							Elem: &schema.Resource{
								Schema: map[string]*schema.Schema{
									"downinter": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"fall": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"guid": {
										Type:     schema.TypeString,
										Computed: true,
									},
									"inter": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"maxconn": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"maxqueue": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"rise": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"slowstart": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
									"weight": {
										Type:     schema.TypeInt,
										Optional: true,
										Computed: true,
									},
								},
							},
						},
					},
				},
			},
		},
	}
}
