/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package rg

import (
	"context"
	"fmt"
	"strconv"

	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/rg"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"

	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
)

func utilityResgroupCheckPresence(ctx context.Context, d *schema.ResourceData, m interface{}) (*rg.RecordRG, error) {
	c := m.(*controller.ControllerCfg)

	idSet := false
	theId, err := strconv.ParseUint(d.Id(), 10, 64)
	if err != nil || theId <= 0 {
		rgId, argSet := d.GetOk("rg_id")
		if argSet {
			theId = uint64(rgId.(int))
			idSet = true
		}
	} else {
		idSet = true
	}

	if idSet {
		log.Debugf("utilityResgroupCheckPresence: locating RG by its ID %d", theId)
		req := rg.GetRequest{
			RGID: theId,
		}

		rgFacts, err := c.CloudBroker().RG().Get(ctx, req)
		if err != nil {
			return nil, err
		}
		return rgFacts, nil
	}

	rgName, argSet := d.GetOk("name")
	if !argSet {
		return nil, fmt.Errorf("Cannot check resource group presence if name is empty and no resource group ID specified")
	}

	listReq := rg.ListRequest{
		IncludeDeleted: false,
	}
	model, err := c.CloudBroker().RG().List(ctx, listReq)
	if err != nil {
		return nil, err
	}

	log.Debugf("utilityResgroupCheckPresence: traversing decoded Json of length %d", len(model))
	for index, item := range model {
		// match by RG name & account ID
		if item.Name == rgName.(string) && item.AccountID == uint64(d.Get("account_id").(int)) {
			log.Debugf("utilityResgroupCheckPresence: match RG name %s / ID %d, account ID %d at index %d",
				item.Name, item.ID, item.AccountID, index)

			req := rg.GetRequest{
				RGID: item.ID,
			}

			apiResp, err := c.CloudBroker().RG().Get(ctx, req)
			if err != nil {
				return nil, err
			}

			return apiResp, nil
		}
	}

	return nil, fmt.Errorf("Cannot find RG name %s owned by account ID %d", rgName, d.Get("account_id").(int))
}
