/*
Copyright (c) 2019-2022 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package sep

import (
	"context"
	"encoding/json"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/sep"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/flattens"
)

func resourceSepCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceSepCreate: called for sep %s", d.Get("name").(string))

	c := m.(*controller.ControllerCfg)
	req := sep.CreateRequest{}

	req.Name = d.Get("name").(string)
	req.GID = uint64(d.Get("gid").(int))
	req.SEPType = d.Get("type").(string)

	if desc, ok := d.GetOk("desc"); ok {
		req.Description = desc.(string)
	}
	if configString, ok := d.GetOk("config"); ok {
		req.Config = configString.(string)
	}
	if enable, ok := d.GetOk("enable"); ok {
		req.Enable = enable.(bool)
	}

	var consumedNIDs []uint64
	for _, item := range d.Get("consumed_by").([]interface{}) {
		consumedNIDs = append(consumedNIDs, uint64(item.(int)))
	}

	req.ConsumerNIDs = consumedNIDs

	var providerNIDs []uint64
	for _, item := range d.Get("provided_by").([]interface{}) {
		providerNIDs = append(providerNIDs, uint64(item.(int)))
	}

	req.ProviderNIDs = providerNIDs

	sepId, err := c.CloudBroker().SEP().Create(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId(strconv.FormatUint(sepId, 10))
	d.Set("sep_id", sepId)

	return resourceSepRead(ctx, d, m)
}

func resourceSepRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceSepRead: called for %s id: %d", d.Get("name").(string), d.Get("sep_id").(int))

	sep, err := utilitySepCheckPresence(ctx, d, m)
	if sep == nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	d.Set("ckey", sep.CKey)
	d.Set("meta", flattens.FlattenMeta(sep.Meta))
	d.Set("consumed_by", sep.ConsumedBy)
	d.Set("desc", sep.Description)
	d.Set("gid", sep.GID)
	d.Set("guid", sep.GUID)
	d.Set("sep_id", sep.ID)
	d.Set("milestones", sep.Milestones)
	d.Set("name", sep.Name)
	d.Set("obj_status", sep.ObjStatus)
	d.Set("provided_by", sep.ProvidedBy)
	d.Set("tech_status", sep.TechStatus)
	d.Set("type", sep.Type)
	data, _ := json.Marshal(sep.Config)
	d.Set("config", string(data))

	return nil
}

func resourceSepDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceSepDelete: called for %s, id: %d", d.Get("name").(string), d.Get("sep_id").(int))

	sepDes, err := utilitySepCheckPresence(ctx, d, m)
	if sepDes == nil {
		if err != nil {
			return diag.FromErr(err)
		}
		return nil
	}

	c := m.(*controller.ControllerCfg)
	req := sep.DeleteRequest{
		SEPID: sepDes.ID,
	}

	_, err = c.CloudBroker().SEP().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceSepUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceSepEdit: called for %s, id: %d", d.Get("name").(string), d.Get("sep_id").(int))
	c := m.(*controller.ControllerCfg)

	if d.HasChange("decommission") {
		decommission := d.Get("decommission").(bool)
		if decommission {
			req := sep.DecommissionRequest{
				SEPID:           uint64(d.Get("sep_id").(int)),
				ClearPhisically: d.Get("clear_physically").(bool),
			}

			_, err := c.CloudBroker().SEP().Decommission(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	if d.HasChange("upd_capacity_limit") {
		updCapacityLimit := d.Get("upd_capacity_limit").(bool)
		if updCapacityLimit {
			req := sep.UpdateCapacityLimitRequest{
				SEPID: uint64(d.Get("sep_id").(int)),
			}

			_, err := c.CloudBroker().SEP().UpdateCapacityLimit(ctx, req)
			if err != nil {
				return diag.FromErr(err)
			}
		}
	}

	if d.HasChange("config") {
		validateReq := sep.ConfigValidateRequest{
			SEPID:  uint64(d.Get("sep_id").(int)),
			Config: d.Get("config").(string),
		}

		insertReq := sep.ConfigInsertRequest{
			SEPID:  uint64(d.Get("sep_id").(int)),
			Config: d.Get("config").(string),
		}

		_, err := c.CloudBroker().SEP().ConfigValidate(ctx, validateReq)
		if err != nil {
			return diag.FromErr(err)
		}
		_, err = c.CloudBroker().SEP().ConfigInsert(ctx, insertReq)
		if err != nil {
			return diag.FromErr(err)
		}

	}

	if d.HasChange("field_edit") {
		fieldConfig := d.Get("field_edit").([]interface{})
		field := fieldConfig[0].(map[string]interface{})
		req := sep.ConfigFieldEditRequest{
			SEPID:      uint64(d.Get("sep_id").(int)),
			FieldName:  field["field_name"].(string),
			FieldValue: field["field_value"].(string),
			FieldType:  field["field_type"].(string),
		}

		_, err := c.CloudBroker().SEP().ConfigFieldEdit(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("enable") {
		err := resourceSepChangeEnabled(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("consumed_by") {
		err := resourceSepUpdateNodes(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("provided_by") {
		err := resourceSepUpdateProviders(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if diagnostics := resourceSepRead(ctx, d, m); diagnostics != nil {
		return diagnostics
	}

	return nil
}

func resourceSepChangeEnabled(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	sepID := uint64(d.Get("sep_id").(int))

	c := m.(*controller.ControllerCfg)
	if d.Get("enable").(bool) {
		req := sep.EnableRequest{
			SEPID: sepID,
		}

		_, err := c.CloudBroker().SEP().Enable(ctx, req)
		if err != nil {
			return err
		}
	} else {
		req := sep.DisableRequest{
			SEPID: sepID,
		}

		_, err := c.CloudBroker().SEP().Disable(ctx, req)
		if err != nil {
			return err
		}
	}

	return nil
}

func resourceSepUpdateNodes(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceSepUpdateNodes: called for %s, id: %d", d.Get("name").(string), d.Get("sep_id").(int))
	c := m.(*controller.ControllerCfg)
	delReq := sep.DelConsumerNodesRequest{
		SEPID: uint64(d.Get("sep_id").(int)),
	}

	addReq := sep.AddConsumerNodesRequest{
		SEPID: uint64(d.Get("sep_id").(int)),
	}

	toDel := false

	t1, t2 := d.GetChange("consumed_by")

	consumedIds := make([]interface{}, 0)

	if d1, d2 := t1.([]interface{}), t2.([]interface{}); len(d1) > len(d2) {
		for _, n := range d2 {
			if !findElInt(d1, n) {
				consumedIds = append(consumedIds, n)
			}
		}
		toDel = true
	} else {
		consumedIds = d.Get("consumed_by").([]interface{})
	}

	var consumerNIDs []uint64
	for _, consumedId := range consumedIds {
		consumerNIDs = append(consumerNIDs, uint64(consumedId.(int)))
	}

	if toDel {
		delReq.ConsumerNIDs = consumerNIDs

		_, err := c.CloudBroker().SEP().DelConsumerNodes(ctx, delReq)
		if err != nil {
			return err
		}
	} else {
		addReq.ConsumerNIDs = consumerNIDs

		_, err := c.CloudBroker().SEP().AddConsumerNodes(ctx, addReq)
		if err != nil {
			return err
		}
	}

	return nil
}

func findElInt(sl []interface{}, el interface{}) bool {
	for _, e := range sl {
		if e.(int) == el.(int) {
			return true
		}
	}
	return false
}

func resourceSepUpdateProviders(ctx context.Context, d *schema.ResourceData, m interface{}) error {
	log.Debugf("resourceSepUpdateProviders: called for %s, id: %d", d.Get("name").(string), d.Get("sep_id").(int))
	c := m.(*controller.ControllerCfg)
	req := sep.AddProviderNodesRequest{
		SEPID: uint64(d.Get("sep_id").(int)),
	}

	var providerNIDs []uint64
	for _, item := range d.Get("provided_by").([]interface{}) {
		providerNIDs = append(providerNIDs, uint64(item.(int)))
	}
	req.ProviderNIDs = providerNIDs

	_, err := c.CloudBroker().SEP().AddProviderNodes(ctx, req)
	if err != nil {
		return err
	}

	return nil
}

func resourceSepSchemaMake() map[string]*schema.Schema {
	return map[string]*schema.Schema{
		"sep_id": {
			Type:        schema.TypeInt,
			Optional:    true,
			Computed:    true,
			Description: "sep type des id",
		},
		"upd_capacity_limit": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "Update SEP capacity limit",
		},
		"decommission": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "unlink everything that exists from SEP",
		},
		"clear_physically": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     true,
			Description: "clear disks and images physically",
		},
		"config": {
			Type:        schema.TypeString,
			Optional:    true,
			Computed:    true,
			Description: "sep config string",
		},
		"ckey": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"meta": {
			Type:     schema.TypeList,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeString,
			},
		},
		"consumed_by": {
			Type:     schema.TypeList,
			Optional: true,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeInt,
			},
			Description: "list of consumer nodes IDs",
		},
		"desc": {
			Type:        schema.TypeString,
			Computed:    true,
			Optional:    true,
			Description: "sep description",
		},
		"gid": {
			Type:        schema.TypeInt,
			Required:    true,
			Description: "grid (platform) ID",
		},
		"guid": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"milestones": {
			Type:     schema.TypeInt,
			Computed: true,
		},
		"name": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "SEP name",
		},
		"obj_status": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"provided_by": {
			Type:     schema.TypeList,
			Optional: true,
			Computed: true,
			Elem: &schema.Schema{
				Type: schema.TypeInt,
			},
			Description: "list of provider nodes IDs",
		},
		"tech_status": {
			Type:     schema.TypeString,
			Computed: true,
		},
		"type": {
			Type:        schema.TypeString,
			Required:    true,
			Description: "type of storage",
		},
		"enable": {
			Type:        schema.TypeBool,
			Optional:    true,
			Default:     false,
			Description: "enable SEP after creation",
		},
		"field_edit": {
			Type:     schema.TypeList,
			MaxItems: 1,
			Optional: true,
			Computed: true,
			Elem: &schema.Resource{
				Schema: map[string]*schema.Schema{
					"field_name": {
						Type:     schema.TypeString,
						Required: true,
					},
					"field_value": {
						Type:     schema.TypeString,
						Required: true,
					},
					"field_type": {
						Type:     schema.TypeString,
						Required: true,
					},
				},
			},
		},
	}
}

func ResourceSep() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceSepCreate,
		ReadContext:   resourceSepRead,
		UpdateContext: resourceSepUpdate,
		DeleteContext: resourceSepDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout60s,
			Read:    &constants.Timeout30s,
			Update:  &constants.Timeout60s,
			Delete:  &constants.Timeout60s,
			Default: &constants.Timeout60s,
		},

		Schema: resourceSepSchemaMake(),
	}
}
